#include <iostream>
#include <fstream>
#include <cudnn.h>
#include <cuda_runtime.h>
#include <sys/time.h>
#include <memory>

static int checkCudaError(cudaError_t code, const char* expr, const char* file, int line) {
    if (code) {
        printf("CUDA error at %s:%d, code=%d (%s) in '%s'", file, line, (int) code, cudaGetErrorString(code), expr);
        return 1;
    }
    return 0;
}

#define checkCudaErr(...)       do { int err = checkCudaError(__VA_ARGS__, #__VA_ARGS__, __FILE__, __LINE__); if (err) return err; } while (0)

static int checkCudnnError(cudnnStatus_t code, const char* expr, const char* file, int line) {
    if (code)  {
        printf("CUDNN error at %s:%d, code=%d (%s) in '%s'\n", file, line, (int) code, cudnnGetErrorString(code), expr);
        return 1;
    }
    return 0;
}

#define checkCudnnErr(...)      do { int err = checkCudnnError(__VA_ARGS__, #__VA_ARGS__, __FILE__, __LINE__);  if (err) return err; } while (0)

int conv_op_process(int& batch, int& in_channel, int& height, int& width, int& kernel_size, int& num_filter, int& stride, int& group, int&iter, bool& half);


int main(int argc, char** argv){
    int batch = 1;
    int in_channel = atoi(argv[1]);
    int height = atoi(argv[2]);
    int width = atoi(argv[3]);
    int num_filter = atoi(argv[4]);
    int kernel_size = atoi(argv[5]);
    int stride = atoi(argv[6]);
    int group = atoi(argv[7]);
    int iter = 1;
    bool half = true;
    printf("%d ", in_channel);
    conv_op_process(batch, in_channel, height, width, kernel_size, num_filter, stride, group, iter, half);
    return 0;
}

using perf_t = cudnnConvolutionFwdAlgoPerf_t;

int read_data(float* data, int size, std::string filename){
    std::ifstream infile(filename);
    if (!infile) {
        std::cerr << "Error: Could not open file." << std::endl;
        return 1;
    }
    int i = 0;
    float num;
    while(infile >> num){
        data[i] = num;
        i++;
    }

    return 0;
}

int conv_op_process(int& batch, int& in_channel, int& height, int& width, int& kernel_size, int& out_channel, int& stride, int& group, int&iter, bool& half){

    int input_size = batch * in_channel * height * width;
    float* A = (float*)malloc(sizeof(float) * input_size);
    int channel_multiplier = out_channel / in_channel;
    int kernel = kernel_size * kernel_size * in_channel * channel_multiplier;
    float* W = (float*)malloc(sizeof(float) * kernel);
    int output_size = batch * out_channel * height * width / stride / stride;
    float* B = (float*)malloc(sizeof(float) * output_size);

    // read input and weight
    int status = read_data(A, input_size, "input.txt");
    status = read_data(W, kernel, "weight.txt");

    // initialize output
    for(int i = 0; i < output_size; i++){
        B[i] = 0;
    }

    // alloc cuda memory
    void* d_A;
    void* d_W;
    void* d_B;
    checkCudaErr(cudaMalloc(&d_A, sizeof(float) * input_size));
    checkCudaErr(cudaMalloc(&d_W, sizeof(float) * kernel));
    checkCudaErr(cudaMalloc(&d_B, sizeof(float) * output_size));

    // mem for warm up
    void* d_A_w;
    void* d_W_w;
    void* d_B_w;
    checkCudaErr(cudaMalloc(&d_A_w, sizeof(float) * input_size));
    checkCudaErr(cudaMalloc(&d_W_w, sizeof(float) * kernel));
    checkCudaErr(cudaMalloc(&d_B_w, sizeof(float) * output_size));

    // memory copy
    checkCudaErr(cudaMemcpy(d_A, (void*)A, sizeof(float) * input_size, cudaMemcpyHostToDevice));
    checkCudaErr(cudaMemcpy(d_W, (void*)W, sizeof(float) * kernel, cudaMemcpyHostToDevice));
    checkCudaErr(cudaMemcpy(d_B, (void*)B, sizeof(float) * output_size, cudaMemcpyHostToDevice));
    checkCudaErr(cudaDeviceSynchronize());    // wait for copy finish

    struct timeval start, handle_create, end;
    // cudnn function init
    cudnnHandle_t handle;
    cudnnTensorDescriptor_t input_desc;
    cudnnTensorDescriptor_t output_desc;
    cudnnFilterDescriptor_t filter_desc;
    cudnnConvolutionDescriptor_t conv_op_desc;
    // handle create process is time cost
    checkCudnnErr(cudnnCreate(&handle));
    gettimeofday(&start, NULL);
    checkCudnnErr(cudnnCreateTensorDescriptor(&input_desc));
    checkCudnnErr(cudnnCreateTensorDescriptor(&output_desc));
    checkCudnnErr(cudnnCreateFilterDescriptor(&filter_desc));
    checkCudnnErr(cudnnCreateConvolutionDescriptor(&conv_op_desc));

    checkCudnnErr(cudnnSetConvolutionGroupCount(conv_op_desc, group));

    cudnnTensorFormat_t infer_type = CUDNN_TENSOR_NHWC;

    int pad = (kernel_size - 1) / 2;

    // input
    checkCudnnErr(cudnnSetTensor4dDescriptor(input_desc, infer_type, CUDNN_DATA_HALF, batch, in_channel, height, width));
    // output
    checkCudnnErr(cudnnSetTensor4dDescriptor(output_desc, infer_type, CUDNN_DATA_HALF, batch, out_channel, int(height/stride), int(width/stride)));
    // filter
    checkCudnnErr(cudnnSetFilter4dDescriptor(filter_desc, CUDNN_DATA_HALF, infer_type, out_channel, int(in_channel/group), kernel_size, kernel_size));

    if(group != 1){
        // CUDNN_CROSS_CORRELATION is very importance in this config
        checkCudnnErr(cudnnSetConvolution2dDescriptor(conv_op_desc, pad, pad, stride, stride, 1, 1, CUDNN_CROSS_CORRELATION, CUDNN_DATA_FLOAT));
    }
    else{
        checkCudnnErr(cudnnSetConvolution2dDescriptor(conv_op_desc, pad, pad, stride, stride, 1, 1, CUDNN_CROSS_CORRELATION, CUDNN_DATA_HALF));
    }
    // using tensor core
    checkCudnnErr( cudnnSetConvolutionMathType(conv_op_desc, CUDNN_TENSOR_OP_MATH));


    printf("conv info %d %d %d %d %d %d %d %d\n", in_channel, height, width, out_channel, kernel_size, pad, stride, group);

    std::unique_ptr<perf_t[]> perf_algos(new perf_t[1]);
    int num_algo = 1;
    int num_algo_get = 0;
    void* workspace_data_temp;
    size_t workspace_data_temp_size = 30 * 1024 * 102;
    checkCudaErr(cudaMalloc(&workspace_data_temp, workspace_data_temp_size));

    cudnnConvolutionFwdAlgo_t algo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM;

    printf("Algo: %d\n", algo);
    size_t workspace_size = 0;
    checkCudnnErr(cudnnGetConvolutionForwardWorkspaceSize(handle, input_desc, filter_desc, conv_op_desc, output_desc, algo, &workspace_size));
    void* workspace;
    cudaMalloc(&workspace, workspace_size);

    // do conv
    float alpha = 1;
    float beta = 0;
    checkCudnnErr(cudnnConvolutionForward(
        handle, (void*)&alpha, input_desc, d_A, filter_desc, d_W,
        conv_op_desc, algo, workspace, workspace_size, (void*)&beta,
        output_desc, d_B));

    printf("------------ CUDNN output ----------\n");
    for(int i = 0; i < output_size; i++){
        printf("%f\t", B[i]);
    }
    printf("\n");

    printf("----------- input --------");
    for(int i = 0; i < input_size; i++){
        printf("%f\t", A[i]);
    }
    printf("\n");

    printf("----------- weight --------");
    for(int i = 0; i < kernel; i++){
        printf("%f\t", W[i]);
    }
    printf("\n");


    // copy data into host
    checkCudaErr(cudaDeviceSynchronize());
    checkCudaErr(cudaMemcpy((void*)B, d_B, sizeof(float) * output_size, cudaMemcpyDeviceToHost));
    // checkCudaErr(cudaMemcpy((void*)A, d_A, sizeof(float) * input_size, cudaMemcpyDeviceToHost));
    cudaDeviceSynchronize();

    return 0;
}

